/* This is an implementation of Algorithm T, explained by Knuth's
   TOACP Section 2.2.3.  It performs a topological sort, given an
   array of "less than" partial ordering relationships.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu) 

This is a preliminary version of it as a Python extension.

*/

#include "Python.h"
#include "PyTopological.h"

static PyObject *PyTopologicalError;


/* I'm surprised that Python's C API doesn't have a calloc defined.
   This is a cheap imitation, but it'll have to do.  */
static void* myPyCalloc(size_t size, int n) {
	void* memory = PyMem_Malloc(size * n);
	if (memory != NULL)    /* Not sure how memset reacts to NULL... */
		memory = memset(memory, 0, size * n);
	return memory;
}


/* Part of the initialization step on Algorithm T involves setting up the
   structure of the Tops nodes.  This function does the gruntwork.  */
static TopNode* getInitializedTops(int relations[], int numPairs, int n) {
	TopNode* nodes = (TopNode*) myPyCalloc(sizeof(TopNode), n + 1);
	SuccessorNode* p;
	int i, j, k;
	for(i = 0; i < numPairs; i++) {
		j = relations[2*i];
		k = relations[2*i+1];
		nodes[k].count++;
		p = (SuccessorNode*) myPyCalloc(sizeof(SuccessorNode), 1);
		p->value = k;
		p->next = nodes[j].top;
		nodes[j].top = p;
	}
	return nodes;
}

/* Deallocating the Tops structure. */
static void freeTops(TopNode* tops, int n) {
	SuccessorNode *p, *q;
	while (n >= 0) {
		p = tops[n].top;    /* Intentional: 'n' doesn't count the sentinel */
		while(p != NULL) {  /* Remove each of the sucessor chains. */
			q = p;
			p = p->next;
			PyMem_Free(q);
		}
		n--;
	}
	PyMem_Free(tops);
}


/* The heart of Algorithm T: perform the topological sort by keeping track of
   dependencies between nodes.  See TOACP Section 2.2.3 for the details about
   this: the cool thing is that it runs REALLY DARN fast, and doesn't use too
   much memory either.  */
static int* algorithmT(int relations[], int num_pairs, int n) {
	TopNode* nodes = getInitializedTops(relations, num_pairs, n);
	int *results = (int*) myPyCalloc(sizeof(int), n);
	SuccessorNode* p;
	int head = 0;
	int tail = 0;
	int i;
	for(i = 1; i <= n; i++) {
		if(nodes[i].count == 0) {
			nodes[tail].count = i;
			tail = i;
		}
	}
	i = 0;
	head = nodes[head].count;
	while(i < n) {
		results[i] = head;

		p = nodes[head].top;
		while(p != NULL) {
			nodes[p->value].count--;
			if (nodes[p->value].count == 0) {
				nodes[tail].count = p->value;
				tail = p->value;
			}
			p = p->next;
		}
		head = nodes[head].count;
		i++;
	}

	freeTops(nodes, n);
	return results;
}




/* ********************************************************************/
/* Up till now was just algorithm implementations.  Now for the really heavy
   Python related stuff... 
*/


/* Returns back a newly allocated array of integers, with the same
   length as that of the original sequence.  If anything bad happens,
   raises TypeError and returns NULL. */
static int* intSequenceToArray(PyObject *seq) {
	int i = 0;
	PyObject *fastseq = NULL;
	int *array = NULL;

	fastseq = PySequence_Fast(seq, "seq must be a sequence.");
	if (fastseq == NULL)
		return NULL;

	array = myPyCalloc(sizeof(int), PySequence_Length(fastseq));
	for(i = 0; i < PySequence_Length(seq); i++) {
		array[i] = PyInt_AsLong(PySequence_Fast_GET_ITEM(fastseq, i));
	}
	Py_DECREF(fastseq);
	return array;
}


/* Inverse of intSequenceToArray: converts an integer array into a PyList. */
static PyObject* intArrayToPyList (int *numbers, int length) {
	PyObject* mylist = PyList_New(length);
	int i;
	for(i = 0; i < PySequence_Length(mylist); i++) {
		PyList_SetItem(mylist, i, PyInt_FromLong(numbers[i]));
	}
	return mylist;
}


/* Small utility function to check up on oddness. */
static int isOdd(int x) { return (x % 2 == 1); }


/* Another function that does a quick sanity check to make sure our
   partial ordering numbers in the array are outside the tolerable
   range.  Used by PyTopological_sortHelper(). */
static int isOutOfRange(int *numbers, int range) {
	int i;
	for(i = 0; i < range; i++) {
		if (numbers[i] < 1 || numbers[i] > range)
			return 1;
	}
	return 0;
}


/* Given a dictionary, returns a new reference that reverses the keys and
   values of the original.  Fairly useful; I'm thinking of making this into
   its own module sometime.  */
static PyObject* invertDict(PyObject *dict) {
	PyObject* inverted = NULL;
	PyObject* items = NULL;
	PyObject* item = NULL;
	int i;

	items = PyDict_Items(dict);
	inverted = PyDict_New();
	for(i = 0; i < PyList_Size(items); i++) {
		item = PyList_GetItem(items, i);
		PyDict_SetItem(inverted, PyTuple_GetItem(item, 1), 
			       PyTuple_GetItem(item, 0));
	}
	Py_DECREF(items);
	return inverted;
}


/* Returns a new reference that translates every element in the list
   with its key value in the dictionary. */
static PyObject* translateList(PyObject *list, PyObject *dict) {
	PyObject* translated_list = NULL;
	PyObject* key = NULL;
	PyObject* value = NULL;
	int i = 0;

	translated_list = PyList_New(PyList_Size(list));
	for(i = 0; i < PyList_Size(list); i++) {
		key = PyList_GetItem(list, i);
		value = PyDict_GetItem(dict, key);
		Py_INCREF(value);
		PyList_SetItem(translated_list, i, value);
	}
	return translated_list;
}


/* Given an element, return a borrowed reference to its enumerated
   value without the dictionary.  If it doesn't exist in dict, add it
   in first. */
static PyObject* getEnumeratedKey(PyObject *dict, PyObject *key) {
	PyObject *value = NULL;
	PyObject *t = NULL;
	value = PyDict_GetItem(dict, key);
	if(value == NULL) {
		t = PyInt_FromLong(PyDict_Size(dict) + 1);
		PyDict_SetItem(dict, key, t);
		Py_DECREF(t);
		value = PyDict_GetItem(dict, key);
	}
	Py_INCREF(value);
	return value;
}


/* Given a list of 2-tuples, returns back a flattened list.  It
   assumes everything is perfect: as a precondition, sequence needs to
   be a sequence.  If anything goes wrong, return NULL.  */
static PyObject* getFlattenedTuplesList(PyObject *sequence) {
	PyObject* flatlist = NULL;
	PyObject* tuple = NULL;
	int i = 0;
	PyObject *t1 = NULL, *t2 = NULL;

	if(! PySequence_Check(sequence))
		return NULL;

	flatlist = PyList_New(PySequence_Length(sequence) * 2);
	for(i = 0; i < PySequence_Length(sequence); i++) {
		tuple = PySequence_GetItem(sequence, i);
		if (!PySequence_Check(tuple) 
		    || PySequence_Length(tuple) != 2) {
			Py_DECREF(tuple);
			Py_DECREF(flatlist);
			return NULL;
		}

		t1 = PySequence_GetItem(tuple, 0);
		t2 = PySequence_GetItem(tuple, 1);
		/* We lend the references to our flatlist. */
		PyList_SetItem(flatlist, 2*i, t1);
		PyList_SetItem(flatlist, 2*i + 1, t2);

		/* Consequently, we should NOT Py_DECREF t1 or t2. */
		Py_DECREF(tuple);

	}
	return flatlist;
}




static char _sort_bad_arg_error[] = 
"partial_orderings should be an even sequence of numbers, all <= n.";

static char PyTopological_sortHelper_doc[] =
"_sort(numbers, n): returns a topological sort of a list of numbers.\n\
\n\
numbers is a flat list, where each pair of numbers represents a partial\n\
ordering.  For example, the sequence:\n\
\n\
    [1, 2, 2, 3]\n\
\n\
represents the partial ordering:\n\
\n\
    1 < 2,\n\
    2 < 3.\n\
\n\
n is number of unique elements in numbers.  An example call to _sort()\n\
might look like:\n\
\n\
    _sort([ 9, 2,   3, 7,   7, 5,   5, 8,\n\
            8, 6,   4, 6,   1, 3,   7, 4,\n\
            9, 5,   2, 8 ], 9)";

/* sort() takes in a sequence of numbers, as well as the number of
   unique elements n.  Each element of the sequence, then, has a
   domain from [1..n].  We treat pairs of elements as "less than"
   relationships.  For example:

   sort([1, 2, 2, 3, 3, 4], 4) asks us to find the topological sort of
   the following:  1 < 2,
                   2 < 3,
		   3 < 4.

   sort() will return back to us a list of those items in topological
   order.  In the case above, we expect to get back the list [1, 2, 3, 4]
*/
static PyObject* PyTopological_sortHelper(PyObject *self, PyObject *args) {
	PyObject *myseq = NULL;
	int unique_items = 0;
	int* numbers = NULL;
	int* sorted_numbers = NULL;
	PyObject *result = NULL;

	if (! PyArg_ParseTuple(args, "Oi", &myseq, &unique_items))
		return NULL; /* Let's make sure we can read things ok. */
  
	numbers = intSequenceToArray(myseq);
	if(numbers == NULL) { return NULL; }

	if(isOdd(PySequence_Length(myseq)) 
	   || isOutOfRange(numbers, PySequence_Length(myseq))) {
		PyErr_SetString(PyTopologicalError, _sort_bad_arg_error);
		PyMem_Free(numbers);
		return NULL;
	}

	sorted_numbers = algorithmT(numbers,
				    PySequence_Length(myseq) / 2,
				    unique_items);
  
	result = intArrayToPyList(sorted_numbers, unique_items);

	PyMem_Free(numbers);
	PyMem_Free(sorted_numbers);
	return result;
}

static const char* bad_format_str = 
"Partial orderings should be a list of 2-tuples.";

static char PyTopological_sort_doc[] = 
"sort(list_of_2_tuples): Given a list of 2-tuple partial\n\
orderings, returns a possible topological sort.\n\
\n\
For example:\n\
\n\
    sort([('underwear', 'pants'),\n\
          ('pants', 'belt'),\n\
          ('pants', 'shoes')])\n\
==> ['underwear', 'pants', 'belt', 'shoes']\n\
\n\
Each element in list_of_2_tuples needs to be a 2-tuple, and hashable.";


/* Wrapper function around PyTopological_sortHelper() which guarantees that we
   call it properly.  It's meant to be easier to use, since it automatically
   handles make the data nice for the helper by renaming elements to numbers.
   After the helper does it works, we translates the results back to their
   original names.

   Note: I know that this would definitely be more reasonable if it were
   written in Python, but I wrote this function to practice with the Python/C
   API, and to understand more about reference counting.  */
static PyObject* PyTopological_sort(PyObject *self, PyObject *args) {
	PyObject* sequence_of_tuples = NULL;
	PyObject* flat_sequence = NULL;
	PyObject* name_map = NULL;
	PyObject* inverted_map = NULL;
	PyObject* mapped_sequence = NULL;
	PyObject* sorted_sequence = NULL;
	PyObject* translated_sequence = NULL;
	PyObject* tmp = NULL;
	int i;

	if (! PyArg_ParseTuple(args, "O", &sequence_of_tuples))
		return NULL;
	flat_sequence = getFlattenedTuplesList(sequence_of_tuples);
	if (! flat_sequence){
		PyErr_SetString(PyTopologicalError, bad_format_str);
		return NULL;
	}

	name_map = PyDict_New();
	mapped_sequence = PyList_New(PyList_Size(flat_sequence));
	for(i = 0; i < PySequence_Length(flat_sequence); i++) {
		tmp = PyList_GetItem(flat_sequence, i);
		PyList_SetItem(mapped_sequence, i,
			       getEnumeratedKey(name_map, tmp));
	}
	inverted_map = invertDict(name_map);

	tmp = Py_BuildValue("Oi", mapped_sequence, PyDict_Size(name_map));
	sorted_sequence = PyTopological_sortHelper(self, tmp);
	Py_DECREF(tmp);

	translated_sequence = translateList(sorted_sequence, inverted_map);

	Py_DECREF(name_map);
	Py_DECREF(inverted_map);
	Py_DECREF(flat_sequence);
	Py_DECREF(mapped_sequence);
	Py_DECREF(sorted_sequence);
	return translated_sequence;
}



static PyMethodDef PyTopologicalMethods[] = {
	{"sort", PyTopological_sort, METH_VARARGS, PyTopological_sort_doc},
	{"_sort", PyTopological_sortHelper, METH_VARARGS,
	 PyTopological_sortHelper_doc},
	{NULL, NULL}
};


static char PyTopological_doc[] = 
"PyTopological provides a topological sorting function for Python.\n\
\n\
It provides the two functions:\n\
\n\
    sort() --- a nicer wrapper around _sort().\n\
    _sort() --- the raw algorithm.\n\
\n\
See their docstrings for more details.\n\
\n\
This is an implementation of Algorithm T, described in Knuth's\n\
\"The Art of Computer Programming\".";
void initPyTopological(void) {
	PyObject *m, *d;
	m = Py_InitModule3("PyTopological", PyTopologicalMethods,
			   PyTopological_doc);
	d = PyModule_GetDict(m);
	PyTopologicalError = PyErr_NewException("PyTopological.error",
						NULL, NULL);
	PyDict_SetItemString(d, "error", PyTopologicalError);
}
